(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Jun Furuse, projet Cristal, INRIA Rocquencourt           *)
(*                                                                     *)
(*  Copyright 1999,2000                                                *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)
open Image

type t = {
    ximage: Gdk.image;
    xwidth: int;
    xheight: int
  } 

let create ~kind ~visual ~width ~height =
  let ximage = Gdk.Image.create ~kind ~visual ~width ~height in
  Gc.finalise (fun t -> Gdk.Image.destroy t) ximage;
  { ximage= ximage;
    xwidth= width;
    xheight= height }
;;

let get drawable ~x ~y ~width ~height =
  let ximage = Gdk.Image.get drawable ~x ~y ~width ~height in
  Gc.finalise (fun t -> Gdk.Image.destroy t) ximage;
  { ximage= ximage;
    xwidth= width;
    xheight= height }
;;

let of_image visual progress img =
  let quick_color_create = Gdk.Truecolor.color_creator visual in
  let quick_color_parser = Gdk.Truecolor.color_parser visual in
  let prog v (* 0.0 .. 1.0 *) = 
    match progress with
    | Some f -> f v
    | None -> ()
  in
  let put_rgb ximg x y rgb =
    let red = rgb.r lsl 8 in
    let green = rgb.g lsl 8 in
    let blue = rgb.b lsl 8 in
    Gdk.Image.put_pixel ximg.ximage ~x ~y 
      ~pixel:(quick_color_create ~red ~green ~blue)
  in
  match img with
  | Rgb24 t ->
      let width = t.Rgb24.width in
      let height = t.Rgb24.height in
      let ximg = create ~kind: `FASTEST ~visual ~width ~height in
      let f_height = float height in
      for y = 0 to height - 1 do
      	for x = 0 to width - 1 do 
	  put_rgb ximg x y (Rgb24.unsafe_get t x y)
	done;
	prog  (float (y+1) /. f_height)
      done;
      ximg

  | Index8 t ->
      let width = t.Index8.width in
      let height = t.Index8.height in
      let cmap = t.Index8.colormap.map in
      let ximg = create ~kind: `FASTEST ~visual ~width ~height in
      let f_height = float height in
      for y = 0 to height - 1 do
      	for x = 0 to width - 1 do 
	  let rgb = 
	    try
	      cmap.(Index8.unsafe_get t x y) 
	    with
	      _ -> {r=255; g=0; b=0} 
	  in
	  put_rgb ximg x y rgb 
	done;
	prog (float (y+1) /. f_height)
      done;
      ximg

  | Index16 t ->
      let width = t.Index16.width in
      let height = t.Index16.height in
      let cmap = t.Index16.colormap.map in
      let ximg = create ~kind: `FASTEST ~visual ~width ~height in
      let f_height = float height in
      for y = 0 to height - 1 do
      	for x = 0 to width - 1 do 
	  let rgb = 
	    try
	      cmap.(Index16.unsafe_get t x y) 
	    with
	      _ -> {r=255; g=0; b=0} 
	  in
	  put_rgb ximg x y rgb 
	done;
	prog (float (y+1) /. f_height)
      done;
      ximg

  | _ -> raise (Failure "not supported")
;;

open GDraw

let get_mono_gc win = 
  let colormap = Gdk.Color.get_system_colormap () in
  let bmp = Gdk.Bitmap.create win ~width:1 ~height: 1 in
  let gc = Gdk.GC.create bmp in
  (* GC.set_foreground gc (Color.color_parse "black"); *)
  Gdk.GC.set_foreground gc (Gdk.Color.alloc ~colormap: colormap `WHITE);
  gc

let plain_mask win w h =
  let colormap = Gdk.Color.get_system_colormap () in
  let mono_gc = get_mono_gc win in
  let bmp = Gdk.Bitmap.create win ~width:w ~height:h in
  Gdk.GC.set_foreground mono_gc (Gdk.Color.alloc ~colormap: colormap `WHITE);
  Gdk.Draw.rectangle bmp mono_gc ~x:0 ~y:0 ~width:w ~height:h ~filled: true ();
  bmp
;;

let mask_of_image win img = (* It is really inefficient *)
  let mono_gc = get_mono_gc win in 
  let width, height = Image.size img in
  let draw_mask t transp image_get =
    prerr_endline "making mask";
    let bmp = Gdk.Bitmap.create win ~width ~height in
    let ximg = get bmp ~x:0 ~y:0 ~width ~height in
    for x = 0 to width - 1 do
      for y = 0 to height - 1 do
	if image_get t x y = transp then
  	  Gdk.Image.put_pixel ximg.ximage ~x ~y ~pixel: 0
	else
  	  Gdk.Image.put_pixel ximg.ximage ~x ~y ~pixel: 1
      done;
    done;
    Gdk.Draw.image bmp mono_gc ~image: ximg.ximage 
      ~xsrc:0 ~ysrc:0 ~xdest:0 ~ydest:0 ~width ~height;
    Some bmp
  in

  (* BUG ? of gtk or lablgtk? Using None for mask does not work *)
  let width, height = Image.size img in
  begin match img with
  | Index8 t ->
      if t.Index8.transparent >= 0 then 
	draw_mask t t.Index8.transparent Index8.unsafe_get 
      else Some (plain_mask win width height)
  | Index16 t ->
      if t.Index16.transparent >= 0 then 
	draw_mask t t.Index16.transparent Index16.unsafe_get 
      else Some (plain_mask win width height)
  | _ -> 
      Some (plain_mask win width height)
  end;
;;

let pixmap_of win ximage =
  let visual = Gdk.Window.get_visual win in
  let pix = Gdk.Pixmap.create win
                          ~width: ximage.xwidth
                          ~height: ximage.xheight
                          ~depth: (Gdk.Visual.depth visual) in
  let pixmap = new drawable pix in
  pixmap#put_image 
    ~x:0 ~y:0 
    ~width: ximage.xwidth ~height: ximage.xheight
    ~xsrc:0 ~ysrc:0 
    ximage.ximage;
  pix
;;

let pixmap_of_image win progress img =
  let visual = Gdk.Window.get_visual win in
  let ximage = of_image visual progress img in
  let msk = mask_of_image win img in
  let pixmap = new GDraw.pixmap ?mask: msk (pixmap_of win ximage) in
  pixmap
;;

